/*
 * YARVCore::VM::Thread
 */

/*
  YARV Thread Desgin

  model 1: Userlevel Thread
    Same as traditional ruby thread.

  model 2: Native Thread with Giant VM lock
    Using pthread (or Windows thread) and Ruby threads run concurrent.

  model 3: Native Thread with fine grain lock
    Using pthread and Ruby threads run concurrent or parallel.

------------------------------------------------------------------------

  model 2:
    A thread has mutex (GVL: Global VM Lock) can run.  When thread
    scheduling, running thread release GVL.  If running thread
    try blocking operation, this thread must release GVL and another
    thread can continue this flow.  After blocking operation, thread
    must check interrupt (CHECK_INTS).

    Every VM can run parallel.

    Ruby threads are scheduled by OS thread scheduler.

------------------------------------------------------------------------

  model 3:
    Every threads run concurrent or parallel and to access shared object
    exclusive access control is needed.  For example, to access String
    object or Array object, fine grain lock must be locked every time.
 */


/* for model 2 */

#include "eval_intern.h"

#define THREAD_DEBUG 0

#if THREAD_DEBUG
#define thread_debug printf
#else
#define thread_debug if(0)printf
#endif

static void sleep_for_polling();
static void sleep_timeval(struct timeval time);
static double timeofday();
struct timeval rb_time_interval(VALUE);

inline static void
st_delete_wrap(st_table *table, VALUE key){
  st_delete(table, (st_data_t *)&key, 0);
}

/********************************************************************************/

#define THREAD_SYSTEM_DEPENDENT_IMPLEMENTATION
#if HAVE_PTHREAD_H
#include "thread_pthread.h"
#elif _WIN32
#include "thread_win32.h"
#else
#error "unsupported thread"
#endif

VALUE th_eval_body(yarv_thread_t *th);

void
static thread_cleanup_func(void *th_ptr)
{
  yarv_thread_t *th = th_ptr;
  st_delete_wrap(GET_VM()->living_threads, th->self);
  th->status = THREAD_KILLED;
  th->machine_stack_start = th->machine_stack_end = 0;
}

static int
thread_start_func_2(yarv_thread_t *th, VALUE *stack_start)
{
  int state;
  VALUE args = th->first_args;
  yarv_proc_t *proc;
  GetProcVal(th->first_proc, proc);

  th->machine_stack_start = stack_start;
  
  thread_debug("start: %p\n", th);
  
  GVL_LOCK_BEGIN();
  {
    yarv_set_current_running_thread(th);
    thread_debug("get_lock, start: %p\n", th);

    TH_PUSH_TAG(th);
    if((state = EXEC_TAG()) == 0){
      th->value = th_invoke_proc(th, proc, RARRAY(args)->len, RARRAY(args)->ptr);
      th->errinfo = Qnil;
    }
    else{
      th->value = Qnil;
    }
    TH_POP_TAG();
    th->status = THREAD_KILLED;
    thread_debug("end: %p\n", th);
    st_delete_wrap(GET_VM()->living_threads, th->self);
  }
  GVL_LOCK_END();

  return 0;
}

static VALUE
yarv_thread_s_new(VALUE klass, VALUE args)
{
  yarv_thread_t *th;
  VALUE thval;
  
  /* create thread object */
  thval= rb_class_new_instance(0, 0, cYarvThread);
  GetThreadVal(thval, th);
  st_insert(GET_VM()->living_threads, thval, (st_data_t)th->thread_id);
  
  /* setup thread environment */
  th->first_args = args;
  th->first_proc = rb_block_proc();

  /* kick thread */
  native_thread_crteate(th);
  return thval;
}

static VALUE
yarv_thread_join(int argc, VALUE *argv, VALUE self)
{
  yarv_thread_t *cur_th = GET_THREAD();
  yarv_thread_t *th;
  int err;

  GetThreadVal(self, th);
  cur_th->wait_thread_value = self;

again:
  if(argc == 0){
    thread_debug("yarv_thread_join(0)\n");

    GVL_UNLOCK_BEGIN();
    {
      err = native_thread_join(th->thread_id, 0);
    }
    GVL_UNLOCK_END();

    if(errno == EINTR){
      goto again;
    }

    switch(err){
    case EDEADLK:
      cur_th->wait_thread_value = Qnil;
      rb_raise(rb_eThreadError, "can't join current thread (cause dead lock)");
    }
  }
  else if(argc == 1){
    double limit = timeofday();
    struct timeval interval;

    interval = rb_time_interval(argv[0]);
    limit += interval.tv_sec;
    limit += interval.tv_usec * 1e-6;

    while(1){
      GVL_UNLOCK_BEGIN();
      {
        sleep_for_polling();
      }
      GVL_UNLOCK_END();

      if(th->status == THREAD_KILLED){
        break;
      }
      if(timeofday() > limit){
        /* timeout */
        cur_th->wait_thread_value = Qnil;
        return self;
      }
    }
  }
  else{
    cur_th->wait_thread_value = Qnil;
    rb_raise(rb_eArgError, "wrong number of arguments");
  }

  cur_th->wait_thread_value = Qnil;
  if(th->errinfo != Qnil){
    rb_exc_raise(th->errinfo);
  }
  return self;
}

static VALUE
yarv_thread_value(VALUE self)
{
  yarv_thread_t *th;

  yarv_thread_join(0, 0, self);
  GetThreadVal(self, th);
  return th->value;
}

#if !defined HAVE_PAUSE
# if defined _WIN32 && !defined __CYGWIN__
#  define pause() Sleep(INFINITE)
# else
#  define pause() sleep(0x7fffffff)
# endif
#endif

void
rb_thread_sleep_forever()
{
  thread_debug("rb_thread_sleep_forever\n");
  GVL_UNLOCK_BEGIN();
  {
    pause();
  }
  GVL_UNLOCK_END();
}

static double
timeofday(void)
{
    struct timeval tv;
    gettimeofday(&tv, NULL);
    return (double)tv.tv_sec + (double)tv.tv_usec * 1e-6;
}

static void
sleep_timeval(struct timeval time)
{
#ifdef _WIN32
  DWORD limit = time.tv_sec * 1000 + time.tv_usec / 1000;
  thread_debug("sleep_timeval - sleep start\n");
  Sleep(limit);
  thread_debug("sleep_timeval - sleep end\n");
  return;
#else
#ifndef linux
  double d, limit;
  limit = timeofday()+(double)time.tv_sec+(double)time.tv_usec*1e-6;
#endif
  thread_debug("sleep_timeval\n");
  for (;;) {
    int n = select(0, 0, 0, 0, &time);
    if (n == 0) return;
    if (n < 0) {
      switch (errno) {
      case EINTR:
#ifdef ERESTART
      case ERESTART:
#endif
        return;
      default:
        rb_sys_fail("sleep");
      }
    }
#ifndef linux
    d = limit - timeofday();

    time.tv_sec = (int)d;
    time.tv_usec = (int)((d - (int)d)*1e6);
    if (time.tv_usec < 0) {
      time.tv_usec += (long)1e6;
      time.tv_sec -= 1;
    }
    if (time.tv_sec < 0) return;
#endif
  }
#endif
}

static void
sleep_for_polling(){
  struct timeval time;
  time.tv_sec  = 0;
  time.tv_usec = 100000; /* 0.1 sec */
  sleep_timeval(time);
}

void
rb_thread_wait_for(struct timeval time)
{
  double date;
  thread_debug("rb_thread_wait_for\n");
  CSL_UNLOCK_BEGIN();
  {
    GVL_UNLOCK_BEGIN();
    {
      sleep_timeval(time);
    }
    GVL_UNLOCK_END();
  }
  CSL_UNLOCK_END();
}

void
yarv_thraed_schedule()
{
  thread_debug("yarv_thraed_schedule\n");
  if(GET_VM()->thread_critical == 0){
    yarv_thread_t *th = GET_THREAD(); 

    thread_debug("yarv_thraed_schedule/switch\n");
    yarv_save_machine_context(th);
    native_mutex_unlock(&GET_VM()->global_interpreter_lock);
    {
      /* do nothing. */
      /* will be switch to another thread which is trying to acquire GVL */
    }
    native_mutex_lock(&GET_VM()->global_interpreter_lock);
    yarv_set_current_running_thread(th);
  }
}

/********************************************************************/

static VALUE
yarv_thread_s_current(VALUE klass){
  return GET_THREAD()->self;
}

static VALUE
yarv_thread_s_main(VALUE klass){
  return GET_VM()->main_thread;
}

VALUE
rb_thread_local_aref(VALUE thread, ID id)
{
  yarv_thread_t *th;
  VALUE val;

  GetThreadVal(thread, th);
  if(ruby_safe_level >= 4 && th != GET_THREAD()){
    rb_raise(rb_eSecurityError, "Insecure: thread locals");
  }
  if(!th->local_storage){
    return Qnil;
  }
  if(st_lookup(th->local_storage, id, &val)){
    return val;
  }
  return Qnil;
}

/*
 *  call-seq:
 *      thr[sym]   => obj or nil
 *  
 *  Attribute Reference---Returns the value of a thread-local variable, using
 *  either a symbol or a string name. If the specified variable does not exist,
 *  returns <code>nil</code>.
 *     
 *     a = Thread.new { Thread.current["name"] = "A"; Thread.stop }
 *     b = Thread.new { Thread.current[:name]  = "B"; Thread.stop }
 *     c = Thread.new { Thread.current["name"] = "C"; Thread.stop }
 *     Thread.list.each {|x| puts "#{x.inspect}: #{x[:name]}" }
 *     
 *  <em>produces:</em>
 *     
 *     #<Thread:0x401b3b3c sleep>: C
 *     #<Thread:0x401b3bc8 sleep>: B
 *     #<Thread:0x401b3c68 sleep>: A
 *     #<Thread:0x401bdf4c run>:
 */

static VALUE
rb_thread_aref(VALUE thread, VALUE id)
{
  return rb_thread_local_aref(thread, rb_to_id(id));
}

VALUE
rb_thread_local_aset(VALUE thread, ID id, VALUE val)
{
  yarv_thread_t *th;
  GetThreadVal(thread, th);
  
  if(ruby_safe_level >= 4 && th != GET_THREAD()){
    rb_raise(rb_eSecurityError, "Insecure: can't modify thread locals");
  }
  if(OBJ_FROZEN(thread)){
    rb_error_frozen("thread locals");
  }
  if(!th->local_storage){
    th->local_storage = st_init_numtable();
  }
  if(NIL_P(val)){
    st_delete(th->local_storage, (st_data_t*)&id, 0);
    return Qnil;
  }
  st_insert(th->local_storage, id, val);
  return val;
}

/*
 *  call-seq:
 *      thr[sym] = obj   => obj
 *  
 *  Attribute Assignment---Sets or creates the value of a thread-local variable,
 *  using either a symbol or a string. See also <code>Thread#[]</code>.
 */

static VALUE
rb_thread_aset(VALUE self, ID id, VALUE val)
{
  return rb_thread_local_aset(self, rb_to_id(id), val);
}

/*
 *  call-seq:
 *     thr.key?(sym)   => true or false
 *  
 *  Returns <code>true</code> if the given string (or symbol) exists as a
 *  thread-local variable.
 *     
 *     me = Thread.current
 *     me[:oliver] = "a"
 *     me.key?(:oliver)    #=> true
 *     me.key?(:stanley)   #=> false
 */

static VALUE
rb_thread_key_p(VALUE self, ID id)
{
  yarv_thread_t *th;
  GetThreadVal(self, th);
  
  if(!th->local_storage){
    return Qfalse;
  }
  if(st_lookup(th->local_storage, rb_to_id(id), 0)){
    return Qtrue;
  }
  return Qfalse;
}

static int
thread_keys_i(ID key, VALUE value, VALUE ary)
{
  rb_ary_push(ary, ID2SYM(key));
  return ST_CONTINUE;
}


/*
 *  call-seq:
 *     thr.keys   => array
 *  
 *  Returns an an array of the names of the thread-local variables (as Symbols).
 *     
 *     thr = Thread.new do
 *       Thread.current[:cat] = 'meow'
 *       Thread.current["dog"] = 'woof'
 *     end
 *     thr.join   #=> #<Thread:0x401b3f10 dead>
 *     thr.keys   #=> [:dog, :cat]
 */

static VALUE
rb_thread_keys(VALUE self)
{
  yarv_thread_t *th;
  VALUE ary = rb_ary_new();
  GetThreadVal(self, th);

  if(th->local_storage){
    st_foreach(th->local_storage, thread_keys_i, ary);
  }
  return ary;
}

static VALUE
yarv_thread_s_pass(VALUE klass)
{
  yarv_thraed_schedule();
  return Qnil;
}

void
yarv_set_stack_end(VALUE **stack_end_p)
{
    VALUE stack_end;
    *stack_end_p = &stack_end;
}

void
yarv_save_machine_context(yarv_thread_t *th)
{
  yarv_set_stack_end(&th->machine_stack_end);
  setjmp(th->machine_regs);
}

static VALUE
raw_gets(VALUE klass){
  char buff[100];
  GVL_UNLOCK_BEGIN();
  {
    fgets(buff, 100, stdin);
  }
  GVL_UNLOCK_END();
  return rb_str_new2(buff);
}

void
Init_yarvthread(){
  rb_define_global_function("raw_gets", raw_gets, 0);
  rb_define_singleton_method(cYarvThread, "new", yarv_thread_s_new, -2);
  rb_define_singleton_method(cYarvThread, "start", yarv_thread_s_new, -2);
  rb_define_singleton_method(cYarvThread, "fork", yarv_thread_s_new, -2);

  rb_define_singleton_method(cYarvThread, "current", yarv_thread_s_current, 0);
  rb_define_singleton_method(cYarvThread, "pass", yarv_thread_s_pass, 0);

  rb_define_method(cYarvThread, "join", yarv_thread_join, -1);
  rb_define_method(cYarvThread, "value", yarv_thread_value, 0);
  rb_define_method(cYarvThread, "[]", rb_thread_aref, 1);
  rb_define_method(cYarvThread, "[]=", rb_thread_aset, 2);
  rb_define_method(cYarvThread, "key?", rb_thread_key_p, 1);
  rb_define_method(cYarvThread, "keys", rb_thread_keys, 0);

  Init_native_thread();
  {
    /* main thread setting */
    {
      /* acquire global interpreter lock */
      yarv_thread_lock_t *lp = &GET_VM()->global_interpreter_lock;
      native_mutex_initialize(lp);
      native_mutex_lock(lp);
    }
  }
}

